//------------------------------------------------------------------------------
// CERN BE-CO-HT
// Mock Turtle
// https://gitlab.cern.ch/coht/mockturtle
//------------------------------------------------------------------------------
//
// unit name:   MQueueHost
//
// description: A SystemVerilog Class to provide an abstraction of the
// MockTurtle core's host message queue.
//
//------------------------------------------------------------------------------
// Copyright CERN 2018
//------------------------------------------------------------------------------
// Copyright and related rights are licensed under the Solderpad Hardware
// License, Version 2.0 (the "License"); you may not use this file except
// in compliance with the License. You may obtain a copy of the License at
// http://solderpad.org/licenses/SHL-2.0.
// Unless required by applicable law or agreed to in writing, software,
// hardware and materials distributed under this License is distributed on an
// "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
// or implied. See the License for the specific language governing permissions
// and limitations under the License.
//------------------------------------------------------------------------------

`ifndef __MT_MQUEUE_HOST_DRIVER_INCLUDED
 `define __MT_MQUEUE_HOST_DRIVER_INCLUDED

 `define MQUEUE_BASE_IN  ('h4000)
 `define MQUEUE_BASE_OUT ('h8000)

 `define MQUEUE_CMD_CLAIM   (1<<24)
 `define MQUEUE_CMD_PURGE   (1<<25)
 `define MQUEUE_CMD_READY   (1<<26)
 `define MQUEUE_CMD_DISCARD (1<<27)

 `define MQUEUE_SLOT_COMMAND 0
 `define MQUEUE_SLOT_STATUS  4

 `define MQUEUE_SLOT_HEADER 16'h1000
 `define MQUEUE_SLOT_DATA   16'h2000

 `define MQUEUE_GCR_SLOT_COUNT   0
 `define MQUEUE_GCR_SLOT_STATUS  4
 `define MQUEUE_GCR_IRQ_MASK     8
 `define MQUEUE_GCR_IRQ_COALESCE 12

 `include "mt_queue_message.svh"

class MQueueHost;
   protected CBusAccessor bus;
   protected uint32_t base;
   protected uint32_t core;
   protected uint32_t slot_count;
   protected MTCPUControl csr;

   typedef MQueueMsg slot_queue_t[$];

   slot_queue_t slots_in[], slots_out[];

   function new( CBusAccessor bus, uint32_t base, uint32_t core,
		 MTCPUControl csr, uint32_t slot_count);
      this.base       = base;
      this.bus        = bus;
      this.core       = core;
      this.csr        = csr;
      this.slot_count = slot_count;
   endfunction // new

   task init();
      this.slots_in  = new[slot_count];
      this.slots_out = new[slot_count];
      for (int i = 0; i < slot_count; i++)
	begin
	   this.slots_in[i]  = {};
	   this.slots_out[i] = {};
	   csr.hmq_select (core, i);
	   incoming_purge ();
	   outgoing_purge ();
	end
   endtask // init

   task read (uint32_t offset, output uint32_t rv);
      uint64_t tmp;
      bus.read (base + offset, tmp);
      rv = tmp;
   endtask // read

   task write (uint32_t offset, uint32_t rv);
      bus.write (base + offset, rv);
   endtask // write

   task get_irq_mask (ref uint32_t val);
      read (`MQUEUE_GCR_IRQ_MASK, val);
   endtask // get_irq_mask

   task set_irq_mask (input uint32_t val);
      write (`MQUEUE_GCR_IRQ_MASK, val);
   endtask // set_irq_mask

   task slot_int_enable(int slot, int enable);
      uint32_t imsk;
      csr.hmq_select ( core, slot );
      get_irq_mask (imsk);
      if (enable)
	imsk |= (1 << slot);
      else
	imsk &= ~(1 << slot);
      set_irq_mask (imsk);
   endtask // slot_int_enable

   task incoming_write (uint32_t r, uint32_t v);
      write (`MQUEUE_BASE_IN + r, v);
   endtask // incoming_write

   task outgoing_write (uint32_t r, uint32_t v);
      write (`MQUEUE_BASE_OUT + r, v);
   endtask // outgoing_write

   task incoming_read (uint32_t r, ref uint32_t v);
      read (`MQUEUE_BASE_IN + r, v);
   endtask // incoming_read

   task outgoing_read (uint32_t r, ref uint32_t v);
      read (`MQUEUE_BASE_OUT + r, v);
   endtask // outgoing_read

   task incoming_purge ();
      incoming_write (`MQUEUE_SLOT_COMMAND, `MQUEUE_CMD_PURGE );
   endtask // incoming_purge

   task outgoing_purge ();
      outgoing_write (`MQUEUE_SLOT_COMMAND, `MQUEUE_CMD_PURGE );
   endtask // outgoing_purge

   task incoming_check_full (output int full);
      uint32_t rv;
      incoming_read (`MQUEUE_SLOT_STATUS, rv);
      full = (rv & 1) ? 1:  0;
   endtask // incoming_check_full

   task outgoing_check_full (output int full);
      uint32_t rv;
      outgoing_read (`MQUEUE_SLOT_STATUS, rv);
      full = (rv & 1) ? 1:  0;
   endtask // outgoing_check_full

   task slot_send ( int slot );

      MQueueMsg msg;

      csr.hmq_select ( core, slot );

      msg = slots_out[slot].pop_back();

      outgoing_write (`MQUEUE_SLOT_COMMAND, `MQUEUE_CMD_CLAIM);
      for (int i = 0; i < msg.header.size(); i++)
	begin
	   outgoing_write (`MQUEUE_SLOT_HEADER + i * 4, msg.header[i]);
	end
      for (int i = 0; i < msg.data.size(); i++)
	begin
	   outgoing_write (`MQUEUE_SLOT_DATA + i * 4, msg.data[i]);
	end
      outgoing_write (`MQUEUE_SLOT_COMMAND, `MQUEUE_CMD_READY);

   endtask // slot_send

   task slot_recv ( int slot );

      MQueueMsg msg;

      uint32_t val;

      csr.hmq_select ( core, slot );

      msg = new (core, slot);

      incoming_read (`MQUEUE_SLOT_HEADER + 0, val );
      msg.header[0] = val;
      incoming_read (`MQUEUE_SLOT_HEADER + 4, val );
      msg.header[1] = val;

      for (int i = 0; i < msg.header[0]; i++)
	begin
	   incoming_read (`MQUEUE_SLOT_DATA + 4*i, val);
	   msg.data.push_back(val);
	end

      incoming_write(`MQUEUE_SLOT_COMMAND, `MQUEUE_CMD_DISCARD );

      slots_in[slot].push_front(msg);

   endtask // slot_recv

   function MQueueMsg receive_message (int slot);
      MQueueMsg ret = slots_in[slot].pop_back();
      return ret;
   endfunction // receive_message

   function void send_message (int slot, MQueueMsg msg);
      slots_out[slot].push_front(msg);
   endfunction // send_message

   function bit pending_in (int slot);
      return slots_in[slot].size() != 0 ? 1'b1 : 1'b0;
   endfunction // pending_in

   task receive_pending();

      uint64_t stat;
      uint32_t core_stat;
      uint32_t imask;

      csr.get_hmq_in_status(stat);

      core_stat = (stat >> (core*8)) & 'hff;

      if (core_stat == 0)
	return;

      csr.hmq_select (core, 0);

      get_irq_mask (imask);

      core_stat &= imask;

      if (core_stat == 0)
	return;

      for (int i = 0; i < slot_count; i++)
	begin
	   csr.hmq_select (core, i);
	   if (core_stat & (1 << i))
	     slot_recv(i);
	end
   endtask // receive_pending

   task send_pending();
      uint32_t full;

      for (int i = 0; i < slot_count ;i++)
	begin
	   while ( slots_out[i].size() )
	     begin
		csr.hmq_select (core, i);
		outgoing_check_full(full);
		if (full)
		  break;
		slot_send(i);
	     end
	end
   endtask // send_pending

endclass // MQueueHost

`endif //  `ifndef __MT_MQUEUE_HOST_DRIVER_INCLUDED
